library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity test_wvt is
    Port ( 
	 	clk : in std_logic;								-- program. clock
--		bar : out std_logic_vector(3 downto 3);	-- 
		sw :	in std_logic_vector(4 downto 1);		-- switches
		lceb, loeb, lweb: out std_logic;		-- SRAM control lines
		la: out std_logic_vector(18 downto 0);	-- SRAM address bus
		ld: inout std_logic_vector(15 downto 0);	-- SRAM data bus
		rceb: out std_logic;		-- SRAM control lines
		rd: out std_logic_vector(9 downto 8)
	 );
end test_wvt;





architecture behavioral of test_wvt is

constant NO  : std_logic := '0';
constant YES : std_logic := '1';
constant LO  : std_logic := '0';
constant HI  : std_logic := '1';

-- Components
--*****************************************************************

-- Component for write test data to RAM
component write_test_data 
    Port ( reset 			: in std_logic;
           clk 			: in std_logic;
           trigger 		: in std_logic;
           data_out 		: out std_logic_vector(7 downto 0);
           adr 			: out std_logic_vector(18 downto 0);
           cen 			: out std_logic;
           oen 			: out std_logic;
           wen 			: out std_logic;
           busy 			: out std_logic);
end component;


-- component for wavelettransformation
component wt_1level
    Port ( resetn : 		in std_logic;
           clk : 			in std_logic;
           trigger : 	in std_logic;
			  horOffset : 	in	 std_logic_vector(9 downto 0); 	--Picture Offset in RAM
			  verOffset : 	in	 std_logic_vector(9 downto 0); 	--Picture Offset in RAM
			  h_size:		in std_logic_vector(9 downto 0);	-- horizontal Picture Size (>16)
			  v_size:		in std_logic_vector(9 downto 0);	-- vertical Picture Size
           pix_adr: 		out std_logic_vector(9 downto 0);
			  line_adr:		out std_logic_vector(9 downto 0);
           data_in : 	in  std_logic_vector(15 downto 0);
           data_out : 	out std_logic_vector(15 downto 0);
			  write_data : out std_logic;
			  cen : 			out std_logic;		-- Chip Enable
			  oen : 			out std_logic;		-- Output Enable
			  wen : 			out std_logic;		-- Write Enable
           busy : 		out std_logic);
end component;

component wt_1LevelUp_8Bit is
    Port ( 		clk 			: in std_logic;
	         	reset 		: in std_logic;
           		trigger 		: in std_logic;
           		h_Offset 	: in std_logic_vector(9 downto 0);
           		v_Offset 	: in std_logic_vector(9 downto 0);
           		h_Size 		: in std_logic_vector(9 downto 0);
           		v_Size 		: in std_logic_vector(9 downto 0);
           		data_in 		: in std_logic_vector(7 downto 0);
           		pix_adr 		: out std_logic_vector(9 downto 0);
           		line_adr 	: out std_logic_vector(9 downto 0);
           		data_out 	: out std_logic_vector(7 downto 0);
           		write_data 	: out std_logic;
           		cen 			: out std_logic;
           		oen 			: out std_logic;
           		wen 			: out std_logic;
           		busy 			: out std_logic);
end component;





component bufg
	port(
		I: in std_logic;
		O: out std_logic
	);
end component;



type STATE_TYPE is (
			IDLE,
			H_TRIG,
			H_TRANS,
			WAIT_STATE1,
			WAIT_STATE2,
			V_TRIG,
			V_TRANS, 
			WAIT_STATE3);


-- signals
--*****************************************************************
signal cnt: std_logic_vector(15 downto 0);
signal slow_clk: std_logic; 		-- clk with appr. 400Hz
signal reset,resetn: std_logic;
signal trigger: std_logic;
signal trigStore1: std_logic_vector(2 downto 0);
signal ce_frl: std_logic;					--chip enable to video ram (active low)
signal oe_frl: std_logic;					--outp enable to video ram (active low)
signal we_frl: std_logic;					--write enable to video ram (active low)
signal adr_frl: std_logic_vector(18 downto 0);					--address into video ram
signal data_frl: std_logic_vector(7 downto 0);				--data into video ram
signal busy_w_testdata: std_logic;		--High when entire fream read
signal gnd_net	: std_logic;		-- logic-0 net
signal LOW, HIGH : std_logic;  -- Constatnt LOW and HIGH
signal iclk,bclk : std_logic;  -- Input and buffered clock
signal mainState: std_logic_vector(2 downto 0);
signal mainState2: std_logic_vector(4 downto 0);

-- for wavelettransform
signal rememberTransform : std_logic;
signal h_size,v_size,verOffset,horOffset : std_logic_vector(9 downto 0);
signal line_adr, pix_adr: std_logic_vector(9 downto 0);
signal data2wvt : std_logic_vector(15 downto 0);
signal data_from_wvt : std_logic_vector(15 downto 0);
signal wr_wvt_data : std_logic;
signal ce_wvt: std_logic;
signal oe_wvt: std_logic;
signal we_wvt: std_logic;
signal busy_wvt,trig_wvt,trig_wvt2: std_logic;
signal trigStore: std_logic_vector(2 downto 0);
signal adr_wvt : std_logic_vector(18 downto 0);
-- for inverse wavelettransform
signal h_size_i,v_size_i,verOffset_i,horOffset_i : std_logic_vector(9 downto 0);
signal line_adr_i, pix_adr_i: std_logic_vector(9 downto 0);
signal data2iwvt : std_logic_vector(7 downto 0);
signal data_from_iwvt : std_logic_vector(7 downto 0);
signal wr_iwvt_data : std_logic;
signal ce_iwvt: std_logic;
signal oe_iwvt: std_logic;
signal we_iwvt: std_logic;
signal busy_iwvt,trig_iwvt,trig_iwvt2: std_logic;
signal trigStoreiwvt: std_logic_vector(2 downto 0);
signal adr_iwvt : std_logic_vector(18 downto 0);



begin

	LOW <= LO;
	HIGH <= HI;


--	ui: ibuf port map(I=>clk, O=>iclk);
--	ub: bufg port map(I=>iclk,O=>bclk);
	bclk <= clk;

	gnd_net <= '0';

	rceb <= HI;		-- disable right-bank of RAM that shares pixel data bus

	reset <= not sw(1);	-- push-button 1 controls reset for the logic circuit
	resetn <= sw(1);

	-- debounce for trigger writeTestData
	process (slow_clk)
	begin
		if slow_clk'event and slow_clk=HI then
			trigStore1(0) <= not sw(2);
			trigStore1(1) <= trigStore1(0);
			trigStore1(2) <= trigStore1(1);
		end if;
	end process;
	trigger <= HI when (trigStore1 = "001") else LO;


	-- debounce for trigger wvt Transform
	process (slow_clk)
	begin
		if slow_clk'event and slow_clk=HI then
			trigStore(0) <= not sw(3);
			trigStore(1) <= trigStore(0);
			trigStore(2) <= trigStore(1);
		end if;
	end process;

	trig_wvt <= HI when (trigStore = "001" and rememberTransform = LO) else LO;

	-- debounce for trigger inv.wvt Transform
	process (slow_clk)
	begin
		if slow_clk'event and slow_clk=HI then
			trigStoreiwvt(0) <= not sw(4);
			trigStoreiwvt(1) <= trigStoreiwvt(0);
			trigStoreiwvt(2) <= trigStoreiwvt(1);
		end if;
	end process;

	trig_iwvt <= HI when (trigStoreiwvt = "001" and rememberTransform = HI) else LO;

	process (reset, clk, busy_wvt, busy_iwvt)
	begin
		if reset = HI then
			rememberTransform <= LO;
		elsif clk'event and clk=HI then
			if busy_wvt=HI then
				rememberTransform <= HI;
			elsif busy_iwvt=HI then
				rememberTransform <= LO;
			else 
				rememberTransform <= rememberTransform;
			end if;
		end if;
	end process;
			

	-- divide the 50 MHz input clock down to 12.5 MHz for the VGA generator
	process(bclk)
	begin
		if (bclk'event and bclk=HI) then
			cnt <= cnt + 1;
		end if;
	end process;

	slow_clk <= cnt(14);

	rd(8)<=bclk;		-- clk for logic Analyzer
	rd(9)<=bclk;		-- clk for logic Analyzer


	mainState <=  (busy_w_testdata & busy_wvt & busy_iwvt);
	mainState2 <= (busy_w_testdata & busy_wvt & wr_wvt_data & busy_iwvt & wr_iwvt_data);

	-- with mainState select source or sink for video ram
	with mainState select
			la(18 downto 0) 	<= adr_frl when  "100",
									   adr_wvt when  "010",
									   adr_iwvt when "001",
									   (others => '0') when others;
	with mainState2 select
			ld(15 downto 0) 	<= ("00000000"&data_frl) when "10000",
										(others => 'Z') when "01000",
										data_from_wvt   when "01100",
										(others => 'Z') when "00010",
										("00000000"&data_from_iwvt)   when "00011",
										(others => 'Z') when others;

	with mainState select
			lceb 					<= ce_frl when "100",
										ce_wvt when "010",
										ce_iwvt when "001",
										'0' when others;

	with mainState select
			loeb					<= oe_frl when  "100",
										oe_wvt when  "010",
										oe_iwvt when "001",
										'1' when others;

	with mainState select
			lweb					<= we_frl when  "100",
										we_wvt when  "010",
										we_iwvt when "001",
										'1' when others;


	with mainState select
			data2wvt				<= ld 					when "010",
										(others => '0') 	when others;

	with mainState select
			data2iwvt			<= ld(7 downto 0)		when "001",
										(others => '0') 	when others;



-- Component for write test data to RAM
wdata: write_test_data 
    Port map ( 
	 		  reset 			=> reset,
           clk 			=> clk,
           trigger 		=> trigger,
           data_out 		=> data_frl,
           adr 			=> adr_frl,
           cen 			=> ce_frl,
           oen 			=> oe_frl,
           wen 			=> we_frl,
           busy 			=> busy_w_testdata
	);


	
	-- Set the Parameters for transformation
	h_size <=    "0000001010";		--10
	v_size <=    "0000000111";		--1
	verOffset <= "0000000001";		--3
	horOffset <= "0000000000";		--0
	adr_wvt <= (line_adr(8 downto 0) & pix_adr(9 downto 0));


	wvt: wt_1level
    port map ( 
	 		  resetn 	=> resetn,
           clk  		=> cnt(2),
           trigger  	=> trig_wvt,
			  horOffset => horOffset,
			  verOffset => verOffset,
			  h_size		=>	h_size,	-- horizontal Picture Size (>16)
			  v_size		=>	v_size,	-- vertical Picture Size
           pix_adr 	=>	pix_adr,
			  line_adr	=>	line_adr,
           data_in	=> data2wvt,
           data_out	=>	data_from_wvt,
			  write_data=> wr_wvt_data,
			  cen			=>	ce_wvt,
			  oen 		=>	oe_wvt,
			  wen 		=>	we_wvt,
           busy 		=> busy_wvt
	);

	-- Set the Parameters for transformation
	h_size_i <=    "0000001010";		--10
	v_size_i <=    "0000000111";		--1
	verOffset_i <= "0000000000";		--3
	horOffset_i <= "0000000000";		--0
	adr_iwvt <= (line_adr_i(8 downto 0) & pix_adr_i(9 downto 0));

	iwvt: wt_1LevelUp_8Bit 
    port map ( clk 			=> cnt(2),
	         	reset 		=> reset,
           		trigger 		=> trig_iwvt,
           		h_Offset 	=> horOffset_i,
           		v_Offset 	=> verOffset_i,
           		h_Size 		=> h_size_i,
           		v_Size 		=> v_size_i,
           		data_in 		=> data2iwvt,
           		pix_adr 		=> pix_adr_i,
           		line_adr 	=> line_adr_i,
           		data_out 	=> data_from_iwvt,
           		write_data 	=> wr_iwvt_data,
           		cen 			=> ce_iwvt,
           		oen 			=> oe_iwvt,
           		wen 			=> we_iwvt,
           		busy 			=> busy_iwvt
	);




end behavioral;
