library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity wt_1LevelUp_8Bit is
    Port ( 		clk 			: in std_logic;
	         	reset 		: in std_logic;
           		trigger 		: in std_logic;
           		h_Offset 	: in std_logic_vector(9 downto 0);
           		v_Offset 	: in std_logic_vector(9 downto 0);
           		h_Size 		: in std_logic_vector(9 downto 0);
           		v_Size 		: in std_logic_vector(9 downto 0);
           		data_in 		: in std_logic_vector(7 downto 0);
           		pix_adr 		: out std_logic_vector(9 downto 0);
           		line_adr 	: out std_logic_vector(9 downto 0);
           		data_out 	: out std_logic_vector(7 downto 0);
           		write_data 	: out std_logic;
           		cen 			: out std_logic;
           		oen 			: out std_logic;
           		wen 			: out std_logic;
           		busy 			: out std_logic);
end wt_1LevelUp_8Bit;

architecture behavioral of wt_1LevelUp_8Bit is

-----------------------------------------------------------------------
-- Components:
-----------------------------------------------------------------------
component cd53_up_8Bit is
	Port ( 	reset 			: in std_logic;	-- active HI reset
				clk 				: in std_logic;
				enablen_in 		: in std_logic;	-- active LO enable
				clk_en_in 		: in std_logic;
				data_in 			: in std_logic_vector(7 downto 0);
				data_out 		: out std_logic_vector(7 downto 0);
				clk_en_out 		: out std_logic;
				enablen_out 	: out std_logic);	-- active LO enable out
end component;

-----------------------------------------------------------------------
-- Constants:
-----------------------------------------------------------------------
constant HI : std_logic := '1';
constant LO : std_logic := '0';
constant YES : std_logic := '1';
constant NO : std_logic := '0';

-- Width of Data
constant BITWIDTH : integer := 7;  -- If 8 Bits are used BITWIDTH must be 7




type STATE_TYPE is (
			IDLE,				
		 	INC_LINE,
		 	WAIT_1,
		 	RD_c0,		
		 	RD_d0,		
			WRITE_1,	
			WRITE_2,	
			GEN_CLKEN);


signal currState, nextState: STATE_TYPE; -- states for main statemachine
signal dataReadyL, dataReadyL2 , dataReadyReset : std_logic;
signal write1,write2: std_logic;
signal readFinished : std_logic;
signal writeFinished : std_logic;
signal pictureFinished : std_logic;
signal data_read : std_logic;
signal read_c0set : std_logic;
signal read_c0reset : std_logic;
signal read_c0 : std_logic;
signal rLine : std_logic_vector(9 downto 0); -- Line Adress Counter
signal wLine : std_logic_vector(9 downto 0); -- Line Adress Counter
signal line_adr2 : std_logic_vector(9 downto 0); -- Line Adress Counter
signal wPix : std_logic_vector(9 downto 0); -- Write Pixel Adress Counter
signal c0Pix : std_logic_vector(9 downto 0); -- Read Pix Approx. Adress Counter
signal d0Pix : std_logic_vector(9 downto 0); -- LRead Pix Detail Adress Counter
signal pix_nr2 : std_logic_vector(9 downto 0); -- Muxed Pixel Counter
signal pix_nr : std_logic_vector(9 downto 0); -- Pixel Counter with Offset added
signal clk_en,clk_en1 :std_logic;

-- iWVT Level 0
signal enable_2_L0 :std_logic;
signal clkEn_f_L0 :std_logic;
signal En_f_L0 :std_logic;
signal data2iwvt :std_logic_vector(BITWIDTH downto 0);
signal data_f_L0 :std_logic_vector(BITWIDTH downto 0);
signal data_f_L0_L1 :std_logic_vector(BITWIDTH downto 0);
signal data_f_L0_L2 :std_logic_vector(BITWIDTH downto 0);


begin




-----------------------------------------------------------------------
-- inverse wavelettransformation Cores
-----------------------------------------------------------------------
wvt_L0: cd53_up_8Bit port map (
			 	reset 			=> reset,			-- active HI reset
				clk 				=> clk,
				enablen_in 		=> enable_2_L0,	-- active LO enable
				clk_en_in 		=> clk_en,
				data_in 			=> data2iwvt,
				data_out 		=> data_f_L0,
				clk_en_out 		=> clkEn_f_L0,
				enablen_out 	=> En_f_L0			-- active LO enable out
	);


-- ====================================================================
-- Detection of Line End and Picture End
-- ====================================================================

-- read Finished Signal
-----------------------------------------------------------------------
	process (d0Pix)
	begin
		if d0Pix >= h_size then
			readFinished <= YES;
		else
			readFinished <= NO;
		end if;
	end process;

-- write Finished Signal
-----------------------------------------------------------------------
	process (readFinished,En_f_L0)
	begin
		if (readFinished = YES and  En_f_L0=HI) then
			writeFinished <= YES;
		else
			writeFinished <= NO;
		end if;
	end process;


-- Picture Finished Signal
-----------------------------------------------------------------------
	process (rLine,readFinished,writeFinished)
	begin
		if (rLine = "0000000000") and readFinished = YES and writeFinished = YES then
			pictureFinished <= YES;
		else
			pictureFinished <= NO;
		end if;
	end process;

-- Enable for Layer 0 iWVT
-----------------------------------------------------------------------
	enable_2_L0 <= readFinished;




-- ====================================================================
-- Latches for data Ready and Read
-- ====================================================================

-- dataReady Signal Latches und iWvtData Latches
--------------------------------------------------------------

-- Latch Data
	process (clk, clkEn_f_L0, dataReadyReset)
	begin
		if clk'event and clk=HI then
			if (dataReadyReset=HI) or (clkEn_f_L0=HI) then
				data_f_L0_L1 <= data_f_L0;
				data_f_L0_L2 <= data_f_L0_L1;
			else
				data_f_L0_L1 <= data_f_L0_L1;
				data_f_L0_L2 <= data_f_L0_L2;
			end if;
		end if;
	end process;

	process (clk, reset, dataReadyReset, clkEn_f_L0)
	begin
		if reset=HI then
			dataReadyL2 <= NO;
			dataReadyL  <= NO;
		elsif clk'event and clk = HI then
			if (dataReadyReset=HI) or (clkEn_f_L0=HI) then
				dataReadyL2 <= clkEn_f_L0;
				dataReadyL  <= dataReadyL2;
			else
				dataReadyL2 <= dataReadyL2;
				dataReadyL  <= dataReadyL;
			end if;
		end if;
	end process;
				


-- read_c0 Signal Set and Reset
--------------------------------------------------------------
	process (read_c0set,read_c0reset)
	begin
		if read_c0set = HI then
			read_c0 <= HI;
		elsif read_c0reset = HI then
			read_c0 <= LO;
		end if;			
	end process;


-- generate clk_enable for L0
--------------------------------------------------------------
	process(clk)
	begin
		if clk'event and clk = HI then
			clk_en1 <= data_read;
		end if;
	end process;
	process(clk)
	begin
		if clk'event and clk = LO then
			clk_en <= clk_en1;
		end if;
	end process;




-- ====================================================================
-- Adresscounters
-- ====================================================================

-- Counter for Lines Start at Lowest Part of Picture; working up
-----------------------------------------------------------------------
	process (clk, reset)
	begin
   	if (reset=HI) then
			rLine <= v_Size - 1;
			wLine <= v_Size;
		elsif (clk'event and clk=HI) then

			if currState=IDLE then 
			-- Linecounter reset after picture is Transformed
				rLine <= v_Size - 1;
				wLine <= v_Size;

			elsif currState=INC_LINE then
			-- Increment lincounter in inc_line State
				rLine <= rLine - 1;
				wLine <= wLine - 1;
			end if;
	  	end if;
	end process;


-- Counter for writePixel
-----------------------------------------------------------------------
	process (clk, reset)
	begin
   	if (reset=HI) then
			wPix <= (others => '0');
		elsif (clk'event and clk=HI) then

			if (currState=INC_LINE) or (currState=IDLE) then 
			-- Pixelcounter reset after line is transformed
				wPix <= (others => '0');

			elsif (currState=WRITE_2) then
			-- Increment pixelcounter after read has been performed
				wPix <= wPix + 1;
			end if;
	  	end if;
	end process;


-- Counter for read c0 Coeff
-----------------------------------------------------------------------
	process (clk, reset)
	begin
   	if (reset=HI) then
			c0Pix <= (others => '0');
		elsif (clk'event and clk=HI) then

			if (currState=INC_LINE) or (currState=IDLE) then 
			-- Pixelcounter reset after line is transformed
				c0Pix <= (others => '0');

			elsif (currState=RD_c0) then
			-- Increment pixelcounter after read has been performed
				c0Pix <= c0Pix + 1;
			end if;
	  	end if;
	end process;


-- Counter for read d0 Coeff
-----------------------------------------------------------------------
	process (clk, reset)
	begin
   	if (reset=HI) then
			d0Pix <= ("0" & h_size(9 downto 1));	-- h_size / 2
		elsif (clk'event and clk=HI) then

			if (currState=INC_LINE) or (currState=IDLE) then 
			-- Pixelcounter reset after line is transformed
				d0Pix <= ("0" & h_size(9 downto 1));	-- h_size / 2

			elsif (currState=RD_d0) then
			-- Increment pixelcounter after read has been performed
				d0Pix <= d0Pix + 1;
			end if;
	  	end if;
	end process;







-- ====================================================================
-- MUX and Latch for Adress (Counters)
-- ====================================================================

	with nextState select
		pix_nr2 <= 	d0Pix		when RD_d0,
						c0Pix		when RD_c0,	
						wPix		when others;

		pix_nr <= h_Offset	+ pix_nr2;

-- Latch the counters at the end of WAIT_1 State
-----------------------------------------------------------------------
	process (clk, currState)
	begin
		if clk'event and clk=HI then
			if currState = WAIT_1 then
				pix_adr <= pix_nr;					-- Pixeladress out
			end if;
		end if;
	end process;


	with currState select							-- LineAdress out
		line_adr2 <= 	rLine when RD_d0,
							rLine when RD_c0,
							wLine when others;

	line_adr <= line_adr2 + v_Offset;


-- ====================================================================
-- RAM Acces Signals
-- ====================================================================
	cen <= LO;	-- RAM always enabled

	with currState select	-- RAM output enable only for read
		oen <= 	LO when RD_d0,
					LO when RD_c0,
					HI when others;

	process(clk)
	begin
		if clk'event and clk=LO then
			wen <= not write1;
		end if;
	end process;

-- data from RAM to iwvt Level 0
-----------------------------------------------------------------------
	process(clk, currState)
	begin
		if clk'event and clk = HI then
			if ((currState = RD_d0) or (currState = RD_c0)) then
			 	data2iwvt <= data_in;
			else
				data2iwvt <= data2iwvt;
			end if;
		end if;
	end process;


-- data from wvt to RAM
-----------------------------------------------------------------------
	with currState select
			write_data <= 	HI when WRITE_1,
								HI when WRITE_2,
								LO when others;


-- Latch the Wavlettransformed Data at the end of WAIT_1 State
-----------------------------------------------------------------------
	process (clk, currState)
	begin
		if clk'event and clk=HI then
			if currState = WAIT_1 then
				data_out <= data_f_L0_L2;
			end if;
		end if;
	end process;





-- ====================================================================
-- ====================================================================
--
-- Statemachine for inverse wavelettransformation 
-- (read, write, send pixel to transform cores)
--
-- ====================================================================
-- ====================================================================
	SYNC_PROC: process (clk, reset)
   begin
   	if (reset=HI) then
      	currState <= IDLE;
      elsif (clk'event and clk = HI) then
         currState <= nextState;
      end if;
   end process;

	COMB_PROC: process (	currState, dataReadyL, dataReadyL2,  trigger,
	 							readFinished, writeFinished,
								pictureFinished)
   begin
		busy <= HI;
		data_read <= LO;
		read_c0set <= LO;
		read_c0reset <= LO;
		dataReadyReset <= LO;
		write1 <= LO;
		write2 <= LO;

		case currState is

		when IDLE =>
			busy <= LO;
			read_c0set <= HI;	-- set read_c0 so that first read is c0
			if trigger=HI then
				nextState <= WAIT_1;
			else
				nextState <= IDLE;
			end if;

		-- Increment Line counter if Picture not Finished
		when INC_LINE =>
			read_c0set <= HI;	-- set read_c0 so that first read is c0
			if pictureFinished=HI then
				nextState <= IDLE;
			else
				nextState <= WAIT_1;
			end if;

		-- State in between of read and write
		when WAIT_1 =>
			if readFinished=HI and writeFinished = HI then
				nextState <= INC_LINE;
			elsif dataReadyL=HI then
				nextState <= WRITE_1;
			elsif (readFinished=LO) and dataReadyL=NO and dataReadyL2=NO and read_c0=YES then
				nextState <= RD_c0;
			elsif (readFinished=LO) and dataReadyL=NO and dataReadyL2=NO and read_c0=NO then 
				nextState <= RD_d0;
			elsif (readFinished=HI) and dataReadyL=NO and dataReadyL2=NO then
				nextState <= GEN_CLKEN;
			else
				nextState <= WAIT_1;
			end if;

		-- read appr. Coeff
		when RD_c0 =>
			data_read <= HI;
			read_c0reset <= HI;
			nextState <= WAIT_1;

		-- read detail coeff.
		when RD_d0 =>
			data_read <= HI;
			read_c0set <= HI;
			nextState <= WAIT_1;

		-- write transformed Data
		when WRITE_1 =>
			write1 <= HI;
			nextState <= WRITE_2;
			
		when WRITE_2 =>
			write2 <= HI;
			dataReadyReset <= HI;
			nextState <= WAIT_1;

		-- generate clk_enable when reading is finished
		when GEN_CLKEN =>
			data_read <= HI;
			nextState <= WAIT_1;

		when others =>
			nextState <= IDLE;
		end case;
	end process;
end behavioral;
